/**********************************************************************************************************************************/
/* Project      : TAPER                                                                                                           */
/* Module       : TPSERPAR.C                                                                                                      */
/* Description  : Serial/Parallel port functions                                                                                  */
/* Version type : Program module                                                                                                  */
/* Last changed : 04-12-1998 10:15                                                                                                */
/* Update count : 5                                                                                                               */
/* OS type      : PC (DOS)                                                                                                        */
/* Compiler     : Watcom C32 Optimizing Compiler Version 10.x                                                                     */
/* Linker       : Watcom C Linker                                                                                                 */
/*                                                                                                                                */
/*                Copyleft (C) 1997-2001 ThunderWare Research Center, written by Martijn van der Heide.                           */
/*                                                                                                                                */
/*                This program is free software; you can redistribute it and/or                                                   */
/*                modify it under the terms of the GNU General Public License                                                     */
/*                as published by the Free Software Foundation; either version 2                                                  */
/*                of the License, or (at your option) any later version.                                                          */
/*                                                                                                                                */
/*                This program is distributed in the hope that it will be useful,                                                 */
/*                but WITHOUT ANY WARRANTY; without even the implied warranty of                                                  */
/*                MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                                                   */
/*                GNU General Public License for more details.                                                                    */
/*                                                                                                                                */
/*                You should have received a copy of the GNU General Public License                                               */
/*                along with this program; if not, write to the Free Software                                                     */
/*                Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                                     */
/*                                                                                                                                */
/**********************************************************************************************************************************/

#define __TPSERPAR_MAIN__

#include "wbench32/wbench32.h"
#include <dos.h>
#include <math.h>
#include "tpconfig.h"
#include "tpaudio.h"
#include "tpserpar.h"

#pragma pack          ()

/**********************************************************************************************************************************/
/* Define the static variables                                                                                                    */
/**********************************************************************************************************************************/

#define THIRTYTWOBITS  (double)4294967295.0

static double     _ParallelTRate = (double)1193180 / (double)3500000;                             /* 1193180 is the PIT frequency */
                                                                                  /* Translates T-states to number of timer ticks */

#define PPT_SPP   0                                                                         /* Standard Parallel Port (old-style) */
#define PPT_PS2   1                                                                           /* PS/2 type (simple bidirectional) */
#define PPT_EPP   2                                                                                     /* Enhanced Parallel Port */
#define PPT_ECP   3                                                                                   /* Enhanced Capability Port */
static byte       _ParallelPortType;

static word       _ParallelPortData;
static word       _ParallelPortStatus;
static word       _ParallelPortControl;
static word       _ParallelPortECPECR;                                                         /* (ECP Extended Control Register) */

static byte       _OldParallelPortControl;
static byte       _OldParallelPortECPECR;

static dword      _ParallelTimer;                                                                                  /* (Bits 0-31) */
static byte       _ParallelTimerHighByte;                                                                         /* (Bits 32-39) */
static double     _RealParallelTimer;
static bool       _JustSetup;

static byte       _OldPIC;

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE INTERNAL LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

static void _SetParallelPort (void)

/**********************************************************************************************************************************/
/* Pre   : This may only be called from Par_PlayInit.                                                                             */
/* Post  : The parallel port type is determined and the port initialized as a simple pass-through SPP.                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  byte Value;

  _ParallelPortType = PPT_SPP;
  _OldParallelPortControl = ReadPort (_ParallelPortControl);
  _OldParallelPortECPECR = ReadPort (_ParallelPortECPECR);                                             /* Check if this is an ECP */
  if ((_OldParallelPortECPECR & 0x03) == 0x01)                                                /* FIFO empty = 1 & FIFO full = 0 ? */
  {
    WritePort (_ParallelPortECPECR, 0x34);                                                 /* Try to set a different control mode */
    Value = ReadPort (_ParallelPortECPECR);                                                                          /* Read back */
    if (Value == 0x35)                                                                  /* Bits 0 and 1 read-only, rest changed ? */
    {
      _ParallelPortType = PPT_ECP;                                                               /* Then we certainly have an ECP */
      WritePort (_ParallelPortECPECR, _OldParallelPortECPECR & 0x1F);             /* Select mode 000 (in bits 5-7, SPP emulation) */
      WritePort (_ParallelPortControl, _OldParallelPortControl & 0x0F);                        /* Disable bidirection and ACK IRQ */
      return;
    }
  }
  WritePort (_ParallelPortControl, _OldParallelPortControl & 0x0F);                            /* Disable bidirection and ACK IRQ */
}

static void _RestoreParallelPort (void)

/**********************************************************************************************************************************/
/* Pre   : _SetParallelPort MUST have been called before (should be called from Par_PlayStop).                                    */
/* Post  : The original parallel port settings have been restored.                                                                */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  if (_ParallelPortType == PPT_ECP)                                                                       /* Did we have an ECP ? */
    WritePort (_ParallelPortECPECR, _OldParallelPortECPECR);                                                  /* Restore the mode */
  WritePort (_ParallelPortControl, _OldParallelPortControl);                                   /* Restore bidirection and ACK IRQ */
}

/**********************************************************************************************************************************/
/* >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> START OF THE LIBRARY ROUTINES <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< */
/**********************************************************************************************************************************/

void Par_PlayInit (void)

/**********************************************************************************************************************************/
/* Pre   : None.                                                                                                                  */
/* Post  : The Parallel port playing variables have been initialized.                                                             */
/* Import: _SetParallelPort.                                                                                                      */
/**********************************************************************************************************************************/

{
  _ParallelPortData    = TPConfig.LPTAddr[TPConfig.LPTPort - 1];                                            /* Get port addresses */
  _ParallelPortStatus  = _ParallelPortData + 1;
  _ParallelPortControl = _ParallelPortData + 2;
  _ParallelPortECPECR  = _ParallelPortData + 0x0402;                                           /* (ECP Extended Control Register) */
  _SetParallelPort ();                                                                 /* Make the port a simple pass-through SPP */
  _ParallelTimer       = ReadTimer ();
  if (TPConfig.ParPlayIRQsDisabled)
  {
    _OldPIC = ReadPort (0x0021);                                                     /* Read hardware interrupt mask from the PIC */
    WritePort (0x0021, 0xF8);                                                              /* Block all hardware IRQs 3 through 7 */
  }
  _JustSetup           = TRUE;
  *SB_DMAOverrun       = FALSE;
  ParallelRunning      = TRUE;                                                                             /* Signal: operational */
}

void Par_PlayStop (void)

/**********************************************************************************************************************************/
/* Pre   : The Parallel port playing is assumed running (but not necessary).                                                      */
/* Post  : The Parallel port playing has been stopped.                                                                            */
/* Import: _RestoreParallelPort.                                                                                                  */
/**********************************************************************************************************************************/

{
  ParallelRunning = FALSE;
  if (TPConfig.ParPlayIRQsDisabled)
    WritePort (0x0021, _OldPIC);                                                         /* Restore all hardware IRQs 3 through 7 */
  _RestoreParallelPort ();                                                                           /* Restore the parallel port */
}

bool Par_PlayPulse (byte State, long PulseLength)

/**********************************************************************************************************************************/
/* Pre   : `State' holds the next state to be played, which will last `PulseLength' T-states.                                     */
/* Post  : The next pulse is started immediately after the current pulse has finished.                                            */
/* Import: None.                                                                                                                  */
/**********************************************************************************************************************************/

{
  static dword _CurrentTimer;
  static byte   NoCacheLoop;

  _CurrentTimer = ReadTimer ();                                                          /* Wait for the previous pulse to finish */

  if (_JustSetup)
  {
    _JustSetup = FALSE;
    _RealParallelTimer = (double)TimerHighByte * THIRTYTWOBITS + (double)_CurrentTimer;                        /* Use all 40 bits */
  }
  else
  {
    if (UsePlayTimeClock)
    {
      while (TimerHighByte < _ParallelTimerHighByte)
      {
        if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (__SBWindow, 1, TEXT_CENTER, PlayTimeString); }
        _CurrentTimer = ReadTimer ();
        NoCacheLoop = TimerHighByte + 1;                 /* (Due to Pentium loop caching, any `TimerHighByte' update is not seen) */
      }
      while (_CurrentTimer < _ParallelTimer)
      {
        if (PlayTimeUpdated)  { PlayTimeUpdated = FALSE; Wprintf (__SBWindow, 1, TEXT_CENTER, PlayTimeString); }
        _CurrentTimer = ReadTimer ();
      }
    }
    else
    {
      while (TimerHighByte < _ParallelTimerHighByte)
      {
        _CurrentTimer = ReadTimer ();
        NoCacheLoop = TimerHighByte + 1;
      }
      while (_CurrentTimer < _ParallelTimer)
        _CurrentTimer = ReadTimer ();
    }
    if (_CurrentTimer - _ParallelTimer > 5000)                                        /* Way too much time passed (about 5 ms.) ? */
    { *SB_DMAOverrun = RealSB_DMAOverrun = TRUE; return (FALSE); }                                    /* Signal: machine too slow */
  }

  _RealParallelTimer += ((double)PulseLength * _ParallelTRate);                             /* Determine clock ticks to next edge */
  _ParallelTimer = (dword)fmod ((_RealParallelTimer + 0.5), THIRTYTWOBITS);                           /* Round to nearest integer */
  _ParallelTimerHighByte = (byte)((_RealParallelTimer + 0.5) / THIRTYTWOBITS);

  WritePort (_ParallelPortData, State >> 7);                                                                   /* Write next edge */

  if (kbhit ())
    if (getch () == 0x1B)
      return (FALSE);
  return (TRUE);
}

#ifdef __SKIP_THIS__
bool Par_Sampler (short WhichWindow)
{
  struct TimerDivider_s
  {
    word                  SampleRate;
    byte                  TimerDelta;
  } TimerDividers[] = { { 44192, 27 },
                        { 39773, 30 },
                        { 35094, 34 },
                        { 30594, 39 },
                        { 25387, 47 },
                        { 20223, 59 } };
  byte NumTimerDividers = sizeof (TimerDeviders) / sizeof (struct TimerDevider_s);
#define MAXSAMPLELENGTH 262144                                /* = 2M samples/block = 47.55 sec at 44.1kHz or 104.86 sec at 20kHz */
  struct SampleChain_s
  {
    struct SampleChain_s  Next;
    byte                  Data[MAXSAMPLELENGTH];
    dword                 LengthInUse;
    byte                  Shifter;
  }                      *SampleChain = NULL;
  struct SampleChain_s   *ChainTail;
  struct SampleChain_s    NewRecord;
  byte                    State;
  byte          register  Shifter;
  byte                   *DataIndex;
  dword                   LengthInUse;
  bool                    Finished = FALSE;
  dword                   TimerRate;
  dword                   NewTimer;
  dword                   CurrentTimer;
  byte                    NoCacheLoop;
  byte                    NumBitsInLastByte;
  word                    BlocksInUse = 0;
  dword                   TotalNumberOfSamples;

  Par_PlayInit ();

  /*** Get SampleRate ***/
  TimerRate = TimerDividers[Speed].TimerDelta;
  SampleRate = TimerDividers[Speed].SampleRate;

  if ((NewRecord = (struct SampleChain_s *)malloc (sizeof (struct SampleChain_s))) == NULL)
    FatalError (1, ErrOutOfMemory, sizeof (struct SampleChain_s));
  BlocksInUse ++;
  NewRecord->Next = NULL;
  ChainTail = SampleChain = NewRecord;
  DataIndex = NewRecord->Data;
  Shifter = 0x80;
  while (!Finished)
  {
    CurrentTimer = ReadTimer ();
    if (_JustSetup)
    {
      _JustSetup = FALSE;
      _ParallelTimer = CurrentTimer;
      _ParallelTimerHighByte = TimerHighByte;
    }
    else
    {
      if (!Shifter)
      {
        if (++ LengthInUse == MAXSAMPLELENGTH)
        {
          ChainTail->LengthInUse = LengthInUse;
          ChainTail->Shifter = Shifter;
          if ((NewRecord = (struct SampleChain_s *)malloc (sizeof (struct SampleChain_s))) == NULL)
          { Finished = TRUE; break; }
          BlocksInUse ++;
          NewRecord->Next = NULL;
          ChainTail->Next = NewRecord;
          ChainTail = NewRecord;
          *(DataIndex = NewRecord->Data) = 0;
          LengthInUse = 0;
        }
        else
          *(++ DataIndex) = 0;
        Shifter = 0x80;
      }
      while (TimerHighByte < _ParallelTimerHighByte)
      {
        CurrentTimer = ReadTimer ();
        NoCacheLoop = TimerHighByte + 1;
      }
      while (CurrentTimer < _ParallelTimer)
        CurrentTimer = ReadTimer ();
      if (CurrentTimer - _ParallelTimer > 5000)                                       /* Way too much time passed (about 5 ms.) ? */
      { *SB_DMAOverrun = RealSB_DMAOverrun = TRUE; break; }                                           /* Signal: machine too slow */

      if (ReadPort (_ParallelPortStatus) & 0x80)                                                              /* BUSY bit reset ? */
        *DataIndex |= Shifter;                                                   /* (Used inverted, but there's no difference...) */
      Shifter >>= 1;
    }

    NewTimer = _ParallelTimer + TimerRate;
    if (NewTimer < _ParallelTimer)                                                                        /* Carry to high byte ? */
      _ParallelTimerHighByte ++;
    _ParallelTimer = NewTimer;

    if (kbhit ())
    {
      getch ();
      Finished = TRUE;
    }
  }
  ChainTail->LengthInUse = LengthInUse;
  ChainTail->Shifter = Shifter;

  Par_PlayStop ();

  NumBitsInLastBytes = 8;
  if (Shifter)
    while ((Shifter >>= 1))
      NumBitsInLastByte --;
  TotalNumberOfSamples = (dword)(BlocksInUse - 1) * (dword)MAXSAMPLELENGTH + ChainTail->LengthInUse + NumBitsInLastByte;
}
#endif
